const C3 = self.C3;

const PLUGIN_INFO = {
  id: "piranha305_queue",
  type: "object",
  hasDomSide: false,
  Acts: {
    "Enqueue": {
          "forward": (inst) => inst.Enqueue,
          
          "autoScriptInterface": true,
          },
"EnqueueWithPiority": {
          "forward": (inst) => inst.EnqueueWithPiority,
          
          "autoScriptInterface": true,
          },
"Dequeue": {
          "forward": (inst) => inst.Dequeue,
          
          "autoScriptInterface": true,
          },
"Clear": {
          "forward": (inst) => inst.Clear,
          
          "autoScriptInterface": true,
          },
"Shuffle": {
          "forward": (inst) => inst.Shuffle,
          
          "autoScriptInterface": true,
          },
"LoadFromJSON": {
          "forward": (inst) => inst.LoadDataFromJSON,
          
          "autoScriptInterface": true,
          }
  },
  Cnds: {
    "IsEmpty": {
          "forward": (inst) => inst.IsEmpty,
          
          "autoScriptInterface": true,
        },
"HasItems": {
          "forward": (inst) => inst.HasItems,
          
          "autoScriptInterface": true,
        },
"OnLastItemDequeued": {
          "forward": (inst) => inst.OnLastItemDequeued,
          
          "autoScriptInterface": true,
        },
"OnItemQueued": {
          "forward": (inst) => inst.OnItemQueued,
          
          "autoScriptInterface": true,
        },
"OnItemDequeued": {
          "forward": (inst) => inst.OnItemDequeued,
          
          "autoScriptInterface": true,
        },
"LoopQueue": {
          "forward": (inst) => inst.LoopQueue,
          
          "autoScriptInterface": true,
        }
  },
  Exps: {
    "LastDequeuedValue": {
          "forward": (inst) => inst.LastDequeuedValue,
          
          "autoScriptInterface": true,
        },
"Size": {
          "forward": (inst) => inst.Size,
          
          "autoScriptInterface": true,
        },
"Peek": {
          "forward": (inst) => inst.Peek,
          
          "autoScriptInterface": true,
        },
"Pop": {
          "forward": (inst) => inst.Pop,
          
          "autoScriptInterface": true,
        },
"PeekLast": {
          "forward": (inst) => inst.PeekLast,
          
          "autoScriptInterface": true,
        },
"LastQueued": {
          "forward": (inst) => inst.LastQueued,
          
          "autoScriptInterface": true,
        },
"AsJSON": {
          "forward": (inst) => inst.AsJSON,
          
          "autoScriptInterface": true,
        },
"LoopItem": {
          "forward": (inst) => inst.LoopItem,
          
          "autoScriptInterface": true,
        }
  },
};

const camelCasedMap = new Map();

function camelCasify(str) {
  // If the string is already camelCased, return it
  if (camelCasedMap.has(str)) {
    return camelCasedMap.get(str);
  }
  // Replace any non-valid JavaScript identifier characters with spaces
  let cleanedStr = str.replace(/[^a-zA-Z0-9$_]/g, " ");

  // Split the string on spaces
  let words = cleanedStr.split(" ").filter(Boolean);

  // Capitalize the first letter of each word except for the first one
  for (let i = 1; i < words.length; i++) {
    words[i] = words[i].charAt(0).toUpperCase() + words[i].substring(1);
  }

  // Join the words back together
  let result = words.join("");

  // If the first character is a number, prepend an underscore
  if (!isNaN(parseInt(result.charAt(0)))) {
    result = "_" + result;
  }

  camelCasedMap.set(str, result);

  return result;
}

const parentClass = {
  object: {
    scripting: self.IInstance,
    instance: C3.SDKInstanceBase,
    plugin: C3.SDKPluginBase,
  },
  world: {
    scripting: self.IWorldInstance,
    instance: C3.SDKWorldInstanceBase,
    plugin: C3.SDKPluginBase,
  },
  dom: {
    scripting: self.IDOMInstance,
    instance: C3.SDKDOMInstanceBase,
    plugin: C3.SDKDOMPluginBase,
  },
};

C3.Plugins[PLUGIN_INFO.id] = class extends (
  parentClass[PLUGIN_INFO.type].plugin
) {
  constructor(opts) {
    if (PLUGIN_INFO.hasDomSide) {
      super(opts, PLUGIN_INFO.id);
    } else {
      super(opts);
    }
  }

  Release() {
    super.Release();
  }
};
const P_C = C3.Plugins[PLUGIN_INFO.id];
P_C.Type = class extends C3.SDKTypeBase {
  constructor(objectClass) {
    super(objectClass);
  }

  Release() {
    super.Release();
  }

  OnCreate() {}
};

//====== SCRIPT INTERFACE ======
const map = new WeakMap();

function getScriptInterface(parentClass, map) {
  return class extends parentClass {
    constructor() {
      super();
      map.set(this, parentClass._GetInitInst().GetSdkInstance());
    }
  };
}


const scriptInterface = getScriptInterface(
  parentClass[PLUGIN_INFO.type].scripting,
  map
);

// extend script interface with plugin actions
Object.keys(PLUGIN_INFO.Acts).forEach((key) => {
  const ace = PLUGIN_INFO.Acts[key];
  if (!ace.autoScriptInterface) return;
  scriptInterface.prototype[camelCasify(key)] = function (...args) {
    const sdkInst = map.get(this);
    P_C.Acts[camelCasify(key)].call(sdkInst, ...args);
  };
});

const addonTriggers = [];

// extend script interface with plugin conditions
Object.keys(PLUGIN_INFO.Cnds).forEach((key) => {
  const ace = PLUGIN_INFO.Cnds[key];
  if (!ace.autoScriptInterface || ace.isStatic || ace.isLooping) return;
  if (ace.isTrigger) {
    scriptInterface.prototype[camelCasify(key)] = function (callback, ...args) {
      const callbackWrapper = () => {
        const sdkInst = map.get(this);
        if (P_C.Cnds[camelCasify(key)].call(sdkInst, ...args)) {
          callback();
        }
      };
      this.addEventListener(key, callbackWrapper, false);
      return () => this.removeEventListener(key, callbackWrapper, false);
    };
  } else {
    scriptInterface.prototype[key] = function (...args) {
      const sdkInst = map.get(this);
      return P_C.Cnds[camelCasify(key)].call(sdkInst, ...args);
    };
  }
});

// extend script interface with plugin expressions
Object.keys(PLUGIN_INFO.Exps).forEach((key) => {
  const ace = PLUGIN_INFO.Exps[key];
  if (!ace.autoScriptInterface) return;
  scriptInterface.prototype[camelCasify(key)] = function (...args) {
    const sdkInst = map.get(this);
    return P_C.Exps[camelCasify(key)].call(sdkInst, ...args);
  };
});
//====== SCRIPT INTERFACE ======

//============ ACES ============
P_C.Acts = {};
P_C.Cnds = {};
P_C.Exps = {};
Object.keys(PLUGIN_INFO.Acts).forEach((key) => {
  const ace = PLUGIN_INFO.Acts[key];
  P_C.Acts[camelCasify(key)] = function (...args) {
    if (ace.forward) ace.forward(this).call(this, ...args);
    else if (ace.handler) ace.handler.call(this, ...args);
  };
});
Object.keys(PLUGIN_INFO.Cnds).forEach((key) => {
  const ace = PLUGIN_INFO.Cnds[key];
  P_C.Cnds[camelCasify(key)] = function (...args) {
    if (ace.forward) return ace.forward(this).call(this, ...args);
    if (ace.handler) return ace.handler.call(this, ...args);
  };
  if (ace.isTrigger && ace.autoScriptInterface) {
    addonTriggers.push({
      method: P_C.Cnds[camelCasify(key)],
      id: key,
    });
  }
});
Object.keys(PLUGIN_INFO.Exps).forEach((key) => {
  const ace = PLUGIN_INFO.Exps[key];
  P_C.Exps[camelCasify(key)] = function (...args) {
    if (ace.forward) return ace.forward(this).call(this, ...args);
    if (ace.handler) return ace.handler.call(this, ...args);
  };
});
//============ ACES ============

function getInstanceJs(parentClass, scriptInterface, addonTriggers, C3) {
  return class extends parentClass {
    constructor(inst, properties) {
      super(inst);

      if (properties) {
      }

      this.items = [];
      this.dequeuedValue = null;
      this.lastItem = null;
      this.lastItemQueued = null;
      this.loopItem = null;
    }

    Enqueue(value) {
      this.EnqueueWithPiority(value, 0);
    }

    EnqueueWithPiority(value, piority) {
      const item = { value: value, piority: piority };
      const index = this.items.findIndex((x) => x.piority < piority);
      if (index === -1) {
        this.items.push(item);
      } else {
        this.items.splice(index, 0, item);
      }
      this.lastItemQueued = item;
      this.Trigger(C3.Plugins.piranha305_queue.Cnds.OnItemQueued);
    }

    Dequeue() {
      if (this.Size() > 0) {
        this.dequeuedValue = this.items.shift().value;
        this.Trigger(C3.Plugins.piranha305_queue.Cnds.OnItemDequeued);

        if (this.IsEmpty()) {
          this.Trigger(C3.Plugins.piranha305_queue.Cnds.OnLastItemDequeued);
        }
      } 
    }

    Shuffle() {
      for (let i = this.Size() - 1; i > 0; i--) {
        const j = Math.floor(Math.random() * (i + 1));
        [this.items[i], this.items[j]] = [this.items[j], this.items[i]];
      }
    }

    Clear() {
      this.items = [];
    }

    IsEmpty() {
      return this.items.length === 0;
    }

    HasItems() {
      return this.items.length > 0;
    }

    OnLastItemDequeued() {
      return true;
    }

    OnItemQueued() {
      return true;
    }

    OnItemDequeued() {
      return true;
    }

    Size() {
      return this.items.length;
    }

    LastDequeuedValue() {
      return this.dequeuedValue;
    }

    Peek() {
      if (this.items.length > 0) {
        return this.items[0].value;
      }
      return -1;
    }

    PeekLast() {
      return this.items[this.items.length - 1].value;
    }

    Pop() {
      this.Dequeue();
      return this.dequeuedValue;
    }

    LastQueued() {
      if (this.lastItemQueued) {
        return this.lastItemQueued.value;
      }
      return -1;
    }

    AsJSON() {
      return JSON.stringify(this.items);
    }

    LoadDataFromJSON(json) {
      this.items = JSON.parse(json);
    }

    _DoForEachTrigger(eventSheetManager, currentEvent, solModifiers, oldFrame, newFrame) {
      eventSheetManager.PushCopySol(solModifiers);
      currentEvent.Retrigger(oldFrame, newFrame);
      eventSheetManager.PopSol(solModifiers)
    }

    LoopQueue(popItems) {
      const runtime = this._runtime;
      const eventSheetManager = runtime.GetEventSheetManager();
      const currentEvent = runtime.GetCurrentEvent();
      const solModifiers = currentEvent.GetSolModifiers();
      const eventStack = runtime.GetEventStack();
      const oldFrame = eventStack.GetCurrentStackFrame();
      const newFrame = eventStack.Push(currentEvent);
      const loopStack = eventSheetManager.GetLoopStack();
      const loop = loopStack.Push();

      runtime.SetDebuggingEnabled(false); 

      for(let i = 0; i < this.items.length; i++) {
        
        this.loopItem = popItems ? this.Pop() : this.items[i].value;
        loop.SetIndex(i);
        this._DoForEachTrigger(eventSheetManager, currentEvent, solModifiers, oldFrame, newFrame);
      }

      runtime.SetDebuggingEnabled(true);
      eventStack.Pop();
      loopStack.Pop();
      return false;
    }

    LoopItem() {
      return this.loopItem;
    }


    Release() {
      super.Release();
    }

    SaveToJson() {
      return {
        items: this.items,
      };
    }

    LoadFromJson(o) {
      this.items = o.items;
    }

    GetDebuggerProperties() {
      return [
        {
          title: "Queue",
          properties: [
            {
              name: "$Items",
              value: this.items.join(", ")
            },
            {
              name: "$Size",
              value: this.items.length,
            },
            {
              name: "$JSON",
              value: JSON.stringify(this.items),
            }
          ],
        },
      ];
    }

    Trigger(method) {
      super.Trigger(method);
      const addonTrigger = addonTriggers.find((x) => x.method === method);
      if (addonTrigger) {
        this.GetScriptInterface().dispatchEvent(new C3.Event(addonTrigger.id));
      }
    }

    GetScriptInterfaceClass() {
      return scriptInterface;
    }
  };
}


P_C.Instance = getInstanceJs(
  parentClass[PLUGIN_INFO.type].instance,
  scriptInterface,
  addonTriggers,
  C3
);
